@@ -2,7 +2,11 @@ import datetime
 import decimal
 import numbers
 from uuid import UUID
-from collections import Mapping, Sequence
+try:
+    from collections.abc import Mapping, Sequence
+except ImportError:
+    # python2
+    from collections import Mapping, Sequence
 
 from fastavro.const import (
     INT_MAX_VALUE, INT_MIN_VALUE, LONG_MAX_VALUE, LONG_MIN_VALUE
@@ -12,8 +16,6 @@ from .schema import extract_record_type, UnknownType, schema_name
 from .six import long, is_str, iterkeys, itervalues
 from ._schema_common import SCHEMA_DEFS
 
-no_value = object()
-
 
 def validate_null(datum, **kwargs):
     """
@@ -257,7 +259,7 @@ def validate_record(datum, schema, parent_ns=None, raise_errors=True):
     _, namespace = schema_name(schema, parent_ns)
     return (
         isinstance(datum, Mapping) and
-        all(validate(datum=datum.get(f['name'], f.get('default', no_value)),
+        all(validate(datum=datum.get(f['name'], f.get('default')),
                      schema=f['type'],
                      field='{}.{}'.format(namespace, f['name']),
                      raise_errors=raise_errors)
