@@ -5,15 +5,15 @@ except ImportError:
 
 from ._schema_common import UnknownType, SchemaParseException
 
-_schema_name = _schema.schema_name
-load_schema = _schema.load_schema
+# Private API
+schema_name = _schema.schema_name
 extract_record_type = _schema.extract_record_type
-populate_schema_defs = _schema.populate_schema_defs
 extract_logical_type = _schema.extract_logical_type
-extract_named_schemas_into_repo = _schema.extract_named_schemas_into_repo
+
+# Public API
+load_schema = _schema.load_schema
+parse_schema = _schema.parse_schema
 
 __all__ = [
-    'UnknownType', 'load_schema', 'extract_record_type',
-    'extract_logical_type', 'extract_named_schemas_into_repo',
-    'SchemaParseException',
+    'UnknownType', 'load_schema', 'SchemaParseException', 'parse_schema',
 ]
