@@ -1,16 +1,15 @@
 try:
     from . import _write
-except ImportError as e:
+except ImportError:
     from . import _write_py as _write
+from . import json_write
 
 # Private API
-SCHEMA_DEFS = _write.SCHEMA_DEFS
-acquaint_schema = _write.acquaint_schema
-WRITERS = _write.WRITERS
 
 # Public API
 writer = _write.writer
 Writer = _write.Writer
+json_writer = json_write.json_writer
 schemaless_writer = _write.schemaless_writer
 LOGICAL_WRITERS = _write.LOGICAL_WRITERS
 
